part of '../ffmpeg.dart';

typedef _DecodeFunc = dynamic Function(Map);

abstract class _IsolateEncodable {
  Map _encode();
}

class IsolateError extends _IsolateEncodable {
  late String message;
  late String stack;
  IsolateError(message, [stack]) {
    if (message is IsolateError) {
      this.message = message.message;
      this.stack = message.stack;
    } else {
      this.message = message.toString();
      this.stack = (stack ?? StackTrace.current).toString();
    }
  }

  @override
  String toString() {
    return stack.isEmpty ? message.toString() : "$message\n$stack";
  }

  static IsolateError? _decode(Map obj) {
    if (obj.containsKey(#jsError))
      return IsolateError(obj[#jsError], obj[#jsErrorStack]);
    return null;
  }

  @override
  Map _encode() {
    return {
      #jsError: message,
      #jsErrorStack: stack,
    };
  }
}

class _IsolateFunction implements _IsolateEncodable {
  int? _isolateId;
  SendPort? _port;
  dynamic _invokable;
  _IsolateFunction._fromId(this._isolateId, this._port);

  _IsolateFunction._new(this._invokable) {
    _handlers.add(this);
  }

  static ReceivePort? _invokeHandler;
  static Set<_IsolateFunction> _handlers = Set();

  static get _handlePort {
    if (_invokeHandler == null) {
      _invokeHandler = ReceivePort();
      _invokeHandler!.listen((msg) async {
        final msgPort = msg[#port];
        try {
          final handler = _handlers.firstWhereOrNull(
            (v) => identityHashCode(v) == msg[#handler],
          );
          if (handler == null) throw Exception('handler released');
          final ret = await handler._handle(msg[#msg]);
          if (msg[#ptr] != null) {
            Pointer<IntPtr>.fromAddress(msg[#ptr]).value = ret;
          }
          if (msgPort != null) {
            msgPort.send(_encodeData(ret));
          }
        } catch (e, stack) {
          final err = _encodeData(IsolateError(e, stack));
          if (msg[#ptr] != null) {
            Pointer<IntPtr>.fromAddress(msg[#ptr]).value = -1;
            print(IsolateError(e, stack));
          }
          if (msgPort != null)
            msgPort.send({
              #error: err,
            });
        }
      });
    }
    return _invokeHandler!.sendPort;
  }

  _send(msg) async {
    if (_port == null) return _handle(msg);
    final evaluatePort = ReceivePort();
    _port!.send({
      #handler: _isolateId,
      #msg: msg,
      #port: evaluatePort.sendPort,
    });
    final result = await evaluatePort.first;
    if (result is Map && result.containsKey(#error))
      throw _decodeData(result[#error], _isolateDecoders);
    return _decodeData(result, _isolateDecoders);
  }

  _destroy() {
    _handlers.remove(this);
  }

  _handle(msg) async {
    switch (msg) {
      case #dup:
        _refCount++;
        return null;
      case #free:
        _refCount--;
        if (_refCount < 0) _destroy();
        return null;
      case #destroy:
        _destroy();
        return null;
    }
    final args = _decodeData(msg[#args], _isolateDecoders);
    return _invokable(args);
  }

  Future call(dynamic positionalArguments) async {
    return _send({
      #args: _encodeData(positionalArguments),
    });
  }

  int callSync(dynamic positionalArguments) {
    if (_port == null)
      return _handle({
        #args: _encodeData(positionalArguments),
      });
    final ptr = malloc<IntPtr>();
    ptr.value = ptr.address;
    _port!.send({
      #handler: _isolateId,
      #msg: {
        #args: _encodeData(positionalArguments),
      },
      #ptr: ptr.address,
    });
    while (ptr.value == ptr.address) sleep(Duration(milliseconds: 1));
    final ret = ptr.value;
    malloc.free(ptr);
    return ret;
  }

  static _IsolateFunction? _decode(Map obj) {
    if (obj.containsKey(#isolateFunctionPort))
      return _IsolateFunction._fromId(
        obj[#isolateFunctionId],
        obj[#isolateFunctionPort],
      );
    return null;
  }

  @override
  Map _encode() {
    return {
      #isolateFunctionId: _isolateId ?? identityHashCode(this),
      #isolateFunctionPort: _port ?? _IsolateFunction._handlePort,
    };
  }

  int _refCount = 0;

  dup() {
    _send(#dup);
  }

  free() {
    _send(#free);
  }

  void destroy() {
    _send(#destroy);
  }
}

class _WrapIsolateProtocolRequest extends ProtocolRequest {
  _IsolateFunction _cb;
  _WrapIsolateProtocolRequest(this._cb, int bufferSize) : super(bufferSize);
  @override
  Future close() => _cb.call({
        #type: #close,
      });

  @override
  int read(Pointer<Uint8> buf, int size) => _cb.callSync({
        #type: #read,
        #buf: buf.address,
        #size: size,
      });

  @override
  int seek(int offset, int whence) => _cb.callSync({
        #type: #seek,
        #offset: offset,
        #whence: whence,
      });
}

abstract class IsolateProtocolRequest extends _IsolateEncodable {
  final int bufferSize;
  Future<int> read(Pointer<Uint8> buf, int size);
  Future<int> seek(int offset, int whence);
  Future<void> closeImpl();
  bool _isCancel = false;
  bool get isCancel => _isCancel;

  Future close() async {
    _isCancel = true;
    await closeImpl();
    __functions?.destroy();
    __functions = null;
  }

  IsolateProtocolRequest(this.bufferSize);

  _IsolateFunction? __functions;
  _IsolateFunction get _functions {
    if (__functions == null)
      __functions = _IsolateFunction._new((msg) async {
        _isCancel = false;
        switch (msg[#type]) {
          case #read:
            return read(
              Pointer.fromAddress(msg[#buf]),
              msg[#size],
            );
          case #seek:
            return seek(
              msg[#offset],
              msg[#whence],
            );
          case #close:
            return close();
        }
      });
    return __functions!;
  }

  @override
  Map _encode() => {
        #urlProtocolBufferSize: bufferSize,
        #urlProtocolCallback: _functions._encode(),
      };

  static ProtocolRequest? _decode(Map obj) {
    if (obj.containsKey(#urlProtocolCallback)) {
      return _WrapIsolateProtocolRequest(
        _IsolateFunction._decode(obj[#urlProtocolCallback])!,
        obj[#urlProtocolBufferSize],
      );
    }

    return null;
  }
}

class IsolatePlayFrame {
  final int? width;
  final int? height;
  final int? timeStamp;
  final Pointer<Uint8>? buffer;
  IsolatePlayFrame._new(this.width, this.height, int? buffer, this.timeStamp)
      : buffer = buffer != null ? Pointer.fromAddress(buffer) : null;

  static IsolatePlayFrame? _decode(Map data) {
    if (data.containsKey(#frameTimeStamp))
      return IsolatePlayFrame._new(
        data[#frameWidth],
        data[#frameHeight],
        data[#frameBuffer],
        data[#frameTimeStamp],
      );
    return null;
  }
}

class IsolateFfmpegStream implements _IsolateEncodable {
  final int? index;
  final int? codecType;
  final int? timeBase;
  IsolateFfmpegStream._new(
    this.index,
    this.codecType,
    this.timeBase,
  );

  @override
  Map _encode() => {
        #streamIndexIsolate: index,
      };

  static IsolateFfmpegStream? _decode(Map data) {
    if (data.containsKey(#streamCodecType))
      return IsolateFfmpegStream._new(
        data[#streamIndex],
        data[#streamCodecType],
        data[#streamTimeBase],
      );
    return null;
  }
}

dynamic _encodeData(data, {Map<dynamic, dynamic>? cache}) {
  if (cache == null) cache = Map();
  if (cache.containsKey(data)) return cache[data];
  if (data is _IsolateEncodable) return data._encode();
  if (data is List) {
    final ret = [];
    cache[data] = ret;
    for (int i = 0; i < data.length; ++i) {
      ret.add(_encodeData(data[i], cache: cache));
    }
    return ret;
  }
  if (data is Map) {
    final ret = {};
    cache[data] = ret;
    for (final entry in data.entries) {
      ret[_encodeData(entry.key, cache: cache)] =
          _encodeData(entry.value, cache: cache);
    }
    return ret;
  }
  return data;
}

dynamic _decodeData(data, decoders, {Map<dynamic, dynamic>? cache}) {
  if (cache == null) cache = Map();
  if (cache.containsKey(data)) return cache[data];
  if (data is List) {
    final ret = [];
    cache[data] = ret;
    for (int i = 0; i < data.length; ++i) {
      ret.add(_decodeData(data[i], decoders, cache: cache));
    }
    return ret;
  }
  if (data is Map) {
    for (final decoder in decoders) {
      final decodeObj = decoder(data);
      if (decodeObj != null) return decodeObj;
    }
    final ret = {};
    cache[data] = ret;
    for (final entry in data.entries) {
      ret[_decodeData(entry.key, decoders, cache: cache)] =
          _decodeData(entry.value, decoders, cache: cache);
    }
    return ret;
  }
  return data;
}

void _runIsolate(Map spawnMessage) async {
  final SendPort sendPort = spawnMessage[#port];
  late final FormatContext ctx;
  try {
    final request = _decodeData(spawnMessage[#request], _isolateDecoders);
    ctx = FormatContext(spawnMessage[#url], request: request);
    ctx.__onFrame = _decodeData(spawnMessage[#onFrame], _isolateDecoders);
  } catch (e, stack) {
    sendPort.send({
      #error: IsolateError(e, stack),
    });
  }
  sendPort.send(_IsolateFunction._new(
    (_msg) => IsolateFormatContext._handleMessage(ctx, _msg),
  )._encode());
}

final _isolateDecoders = <_DecodeFunc>[
  IsolateFfmpegStream._decode,
  _IsolateFunction._decode,
  IsolatePlayFrame._decode,
  IsolateError._decode,
  IsolateProtocolRequest._decode,
];

class IsolateFormatContext {
  Future<_IsolateFunction>? _isolate;
  IsolateProtocolRequest? request;
  final void Function(IsolatePlayFrame frame)? onFrame;
  _IsolateFunction? _onFrameIsolate;

  final String url;

  IsolateFormatContext(this.url, {this.onFrame, this.request});

  _ensureEngine() {
    if (_isolate != null) return;
    if (_onFrameIsolate == null && onFrame != null) {
      _onFrameIsolate = _IsolateFunction._new(onFrame);
    }
    ReceivePort port = ReceivePort();
    Isolate.spawn(
      _runIsolate,
      {
        #port: port.sendPort,
        #url: url,
        #onFrame: _onFrameIsolate?._encode(),
        #request: request?._encode(),
      },
      errorsAreFatal: true,
    );
    _isolate = port.first.then((result) {
      port.close();
      if (result is Map && result.containsKey(#error))
        throw _decodeData(result[#error], _isolateDecoders);
      return _decodeData(result, _isolateDecoders) as _IsolateFunction;
    });
  }

  Future<bool> isPlaying() async {
    _ensureEngine();
    return await (await _isolate!)({
      #type: #isPlaying,
    });
  }

  Future<int> getPosition() async {
    _ensureEngine();
    return await (await _isolate!)({
      #type: #getPosition,
    });
  }

  Future<int> duration() async {
    _ensureEngine();
    return await (await _isolate!)({
      #type: #duration,
    });
  }

  Future play(
    List<IsolateFfmpegStream> streams,
  ) async {
    _ensureEngine();
    return (await _isolate!)({
      #type: #play,
      #streams: streams,
    });
  }

  Future pause() async {
    _ensureEngine();
    return (await _isolate!)({
      #type: #pause,
    });
  }

  Future seekTo(
    int ts, {
    int streamIndex = -1,
    int minTs = INT64_MIN,
    int maxTs = INT64_MAX,
    int flags = 0,
  }) async {
    request?._isCancel = true;
    _ensureEngine();
    return (await _isolate!)({
      #type: #seekTo,
      #ts: ts,
      #streamIndex: streamIndex,
      #minTs: minTs,
      #maxTs: maxTs,
      #flags: flags,
    });
  }

  Future resume() async {
    _ensureEngine();
    return (await _isolate!)({
      #type: #resume,
    });
  }

  Future<List<IsolateFfmpegStream>> getStreams() async {
    _ensureEngine();
    return List<IsolateFfmpegStream>.from(await (await _isolate!)({
      #type: #getStreams,
    }));
  }

  close() {
    if (_isolate == null) return;
    final ret = _isolate?.then((isolate) async {
      final closePort = ReceivePort();
      await isolate({
        #type: #close,
        #port: closePort.sendPort,
      });
      isolate.destroy();
      _onFrameIsolate?.destroy();
      _onFrameIsolate = null;
      request?.close();
      request = null;
    });
    _isolate = null;
    return ret;
  }

  static _handleMessage(FormatContext ctx, dynamic _msg) async {
    final decoders = <_DecodeFunc>[
      (obj) {
        final index = obj[#streamIndexIsolate];
        if (index != null) return ctx._streams![index];
        return null;
      }
    ];
    final msg = _decodeData(
      _msg,
      decoders,
    );
    switch (msg[#type]) {
      case #getStreams:
        return ctx.getStreams();
      case #play:
        return ctx.play(List<FfmpegStream>.from(msg[#streams]));
      case #pause:
        return ctx.pause();
      case #resume:
        return ctx.resume();
      case #isPlaying:
        return ctx.isPlaying();
      case #getPosition:
        return ctx.getPosition();
      case #duration:
        return ctx.duration();
      case #seekTo:
        return ctx.seekTo(
          msg[#ts],
          streamIndex: msg[#streamIndex],
          minTs: msg[#minTs],
          maxTs: msg[#maxTs],
          flags: msg[#flags],
        );
      case #close:
        await ctx.close();
        return _IsolateFunction._invokeHandler?.close();
    }
  }
}
